import abc
import asyncio
import datetime
from typing import List, Optional, Tuple, Union
from sqlalchemy.orm import Session, Query
from sqlalchemy.ext.asyncio import AsyncSession
from missionpanel.orm import Mission, Tag, MissionTag, Attempt
from sqlalchemy import select, func, distinct, Select, cast, Text


class HandlerInterface(abc.ABC):

    @staticmethod
    def query_missions_by_tag(tags: List[str]) -> Select[Tuple[Mission]]:
        return (
            select(Mission)
            .join(MissionTag)
            .join(Tag)
            .filter(Tag.name.in_(tags))
            .group_by(Mission.id)
            .having(func.count(distinct(Tag.name)) == len(tags))
        )

    @staticmethod
    def query_todo_missions(tags: List[str]) -> Select[Tuple[Mission]]:
        return (
            HandlerInterface.query_missions_by_tag(tags)
            .outerjoin(Attempt, onclause=(
                (cast(Attempt.content, Text) == cast(Mission.content, Text)) & (
                    # see if Attempt is finished or working on the Mission
                    Attempt.success.is_(True) |  # have finished handler
                    (Attempt.last_update_time + Attempt.max_time_interval >= datetime.datetime.now())  # have working handler
                )
            ))
            .where(Attempt.id == None)
        )

    @staticmethod
    def create_attempt(session: Union[Session | AsyncSession], mission: Mission, name: str, max_time_interval: datetime.timedelta = datetime.timedelta(seconds=1)) -> Attempt:
        attempt = Attempt(
            handler=name,
            max_time_interval=max_time_interval,
            content=mission.content,
            mission=mission)
        session.add(attempt)
        return attempt


class Handler(HandlerInterface, abc.ABC):
    def __init__(self, session: Session, name: str, max_time_interval: datetime.timedelta = datetime.timedelta(seconds=1)):
        self.session = session
        self.name = name
        self.max_time_interval = max_time_interval

    @abc.abstractmethod
    def select_mission(self, missions: Query[Mission]) -> Optional[Mission]:
        return missions[0] if missions else None

    def report_attempt(self, mission: Mission, attempt: Attempt):
        attempt.success = False
        attempt.last_update_time = datetime.datetime.now()
        self.session.commit()

    @abc.abstractmethod
    def execute_mission(self, mission: Mission, attempt: Attempt) -> bool:
        pass

    def run_once(self, tags: List[str]):
        missions = self.session.execute(HandlerInterface.query_todo_missions(tags)).scalars().all()
        mission = self.select_mission(missions)
        if mission is None:
            return
        attempt = HandlerInterface.create_attempt(self.session, mission, self.name, self.max_time_interval)
        self.report_attempt(mission, attempt)
        if self.execute_mission(mission, attempt):
            attempt.success = True
        self.report_attempt(mission, attempt)
        return attempt


class AsyncHandler(HandlerInterface, abc.ABC):
    def __init__(self, session: AsyncSession, name: str, max_time_interval: datetime.timedelta = datetime.timedelta(seconds=1)):
        self.session = session
        self.name = name
        self.max_time_interval = max_time_interval

    @abc.abstractmethod
    async def select_mission(self, missions: Query[Mission]) -> Optional[Mission]:
        return missions[0] if missions else None

    async def report_attempt(self, mission: Mission, attempt: Attempt):
        attempt.last_update_time = datetime.datetime.now()
        await self.session.commit()
        await self.session.refresh(attempt)
        await self.session.refresh(mission)

    @abc.abstractmethod
    async def execute_mission(self, mission: Mission, attempt: Attempt) -> bool:
        pass

    async def get_mission(self, tags: List[str]) -> Optional[Mission]:
        missions = (await self.session.execute(HandlerInterface.query_todo_missions(tags))).scalars().all()
        return await self.select_mission(missions)

    async def watchdog_mission(self, mission: Mission, attempt: Attempt):
        await self.report_attempt(mission, attempt)
        task = asyncio.create_task(self.execute_mission(mission, attempt))
        while not task.done():
            await self.report_attempt(mission, attempt)
            await asyncio.sleep(self.max_time_interval.total_seconds() / 2)
        if task.result():
            attempt.success = True
        await self.report_attempt(mission, attempt)
        return attempt

    async def run_once(self, tags: List[str]):
        mission = await self.get_mission(tags)
        if mission is None:
            return
        attempt = HandlerInterface.create_attempt(self.session, mission, self.name, self.max_time_interval)
        return await self.watchdog_mission(mission, attempt)

    async def run_all(self, tags: List[str]):
        while await self.run_once(tags):
            pass



from typing import List, Union
from sqlalchemy.ext.asyncio import AsyncSession
from missionpanel.orm import Mission
from .abc import SubmitterInterface


class AsyncSubmitterInterface(SubmitterInterface):

    @staticmethod
    async def _query_mission(session: AsyncSession, match_patterns: List[str]) -> Mission:
        matcher = (await session.execute(SubmitterInterface.query_matcher(match_patterns))).scalars().first()
        if matcher is None:
            return None
        mission = await matcher.awaitable_attrs.mission
        existing_matchers = await mission.awaitable_attrs.matchers
        SubmitterInterface.add_mission_matchers(session, mission, match_patterns, existing_matchers)
        return mission

    @staticmethod
    async def _add_tags(session: AsyncSession, mission: Union[Mission | None] = None, tags: List[str] = []):
        exist_tags = (await session.execute(SubmitterInterface.query_tag(tags))).scalars().all() if len(tags) > 0 else []
        exist_mission_tags = await mission.awaitable_attrs.tags
        SubmitterInterface.add_mission_tags(session, mission, tags, exist_tags, exist_mission_tags)

    @staticmethod
    async def match_mission(session: AsyncSession, match_patterns: List[str]) -> Mission:
        mission = await AsyncSubmitterInterface._query_mission(session, match_patterns)
        await session.commit()
        return mission

    @staticmethod
    async def create_mission(session: AsyncSession, content: str, match_patterns: List[str], tags: List[str] = []):
        mission = await AsyncSubmitterInterface._query_mission(session, match_patterns)
        mission = SubmitterInterface.create_mission(session, content, match_patterns, mission)
        await AsyncSubmitterInterface._add_tags(session, mission, tags)
        await session.commit()
        await session.refresh(mission)
        return mission

    @staticmethod
    async def add_tags(session: AsyncSession, match_patterns: List[str], tags: List[str]):
        mission = await AsyncSubmitterInterface._query_mission(session, match_patterns)
        if mission is None:
            raise ValueError("Mission not found")
        await AsyncSubmitterInterface._add_tags(session, mission, tags)
        await session.commit()


class AsyncSubmitter(AsyncSubmitterInterface):
    def __init__(self, session: AsyncSession):
        self.session = session

    async def match_mission(self, match_patterns: List[str]) -> Mission:
        return await AsyncSubmitterInterface.match_mission(self.session, match_patterns)

    async def create_mission(self, content: str, match_patterns: List[str]):
        return await AsyncSubmitterInterface.create_mission(self.session, content, match_patterns)

    async def add_tags(self, matchers: List[str], tags: List[str]):
        return await AsyncSubmitterInterface.add_tags(self.session, matchers, tags)



from typing import List, Union
from sqlalchemy.orm import Session
from missionpanel.orm import Mission
from .abc import SubmitterInterface


class SyncSubmitterInterface(SubmitterInterface):

    @staticmethod
    def _query_mission(session: Session, match_patterns: List[str]) -> Mission:
        matcher = session.execute(SubmitterInterface.query_matcher(match_patterns)).scalars().first()
        if matcher is None:
            return None
        SubmitterInterface.add_mission_matchers(session, matcher.mission, match_patterns, matcher.mission.matchers)
        return matcher.mission

    @staticmethod
    def _add_tags(session: Session, mission: Union[Mission | None] = None, tags: List[str] = []):
        exist_tags = session.execute(SubmitterInterface.query_tag(tags)).scalars().all()
        tags = SubmitterInterface.add_mission_tags(session, mission, tags, exist_tags, mission.tags)

    @staticmethod
    def match_mission(session: Session, match_patterns: List[str]) -> Mission:
        mission = SyncSubmitterInterface._query_mission(session, match_patterns)
        session.commit()
        return mission

    @staticmethod
    def create_mission(session: Session, content: str, match_patterns: List[str], tags: List[str] = []):
        mission = SyncSubmitterInterface._query_mission(session, match_patterns)
        mission = SubmitterInterface.create_mission(session, content, match_patterns, mission)
        SyncSubmitterInterface._add_tags(session, mission, tags)
        session.commit()
        return mission

    @staticmethod
    def add_tags(session: Session, match_patterns: List[str], tags: List[str]):
        mission = SyncSubmitterInterface._query_mission(session, match_patterns)
        if mission is None:
            raise ValueError("Mission not found")
        SyncSubmitterInterface._add_tags(session, mission, tags)
        session.commit()


class Submitter(SyncSubmitterInterface):
    def __init__(self, session: Session):
        self.session = session

    def match_mission(self, match_patterns: List[str]) -> Mission:
        return SyncSubmitterInterface.match_mission(self.session, match_patterns)

    def create_mission(self, content: str, match_patterns: List[str], tags: List[str] = []):
        return SyncSubmitterInterface.create_mission(self.session, content, match_patterns, tags)

    def add_tags(self, match_patterns: List[str], tags: List[str]):
        return SyncSubmitterInterface.add_tags(self.session, match_patterns, tags)



from typing import List, Union, Tuple
from sqlalchemy.orm import Session
from sqlalchemy.ext.asyncio import AsyncSession
from missionpanel.orm import Mission, Tag, Matcher, MissionTag
from sqlalchemy import select, Select


class SubmitterInterface:
    '''
    SubmitterInterface is an interface for Submitter and AsyncSubmitter to implement the common methods.
    There is no anything about session.execute in this class.
    '''
    @staticmethod
    def query_matcher(match_patterns: List[str]) -> Select[Tuple[Matcher]]:
        return select(Matcher).where(Matcher.pattern.in_(match_patterns)).limit(1).with_for_update()

    @staticmethod
    def add_mission_matchers(session: Union[Session | AsyncSession], mission: Mission, match_patterns: List[str], existing_matchers: List[Matcher] = []) -> Mission:
        exist_patterns = [matcher.pattern for matcher in existing_matchers]
        session.add_all([Matcher(pattern=pattern, mission=mission) for pattern in match_patterns if pattern not in exist_patterns])

    @staticmethod
    def create_mission(session: Union[Session | AsyncSession], content: str, match_patterns: List[str], existing_mission: Union[Mission | None] = None) -> Mission:
        if existing_mission is None:
            mission = Mission(
                content=content,
                matchers=[Matcher(pattern=pattern) for pattern in match_patterns],
            )
            session.add(mission)
        else:
            mission = existing_mission
            if mission.content != content:
                mission.content = content
        return mission

    @staticmethod
    def query_tag(tags_name: List[str]) -> Select[Tuple[Matcher]]:
        return select(Tag).where(Tag.name.in_(tags_name)).with_for_update()

    @staticmethod
    def add_mission_tags(session: Union[Session | AsyncSession], mission: Mission, tags_name: List[str], exist_tags: List[Tag] = [], exist_mission_tags: List[MissionTag] = []):
        exist_tags_name = [tag.name for tag in exist_tags]
        session.add_all([Tag(name=tag_name) for tag_name in tags_name if tag_name not in exist_tags_name])
        exist_mission_tags_name = [tag.tag_name for tag in exist_mission_tags]
        session.add_all([MissionTag(mission=mission, tag_name=tag_name) for tag_name in tags_name if tag_name not in exist_mission_tags_name])



import abc
import traceback
from typing import AsyncGenerator, List, Any
import httpx
import logging
from xml.etree import ElementTree
from missionpanel.submitter import AsyncSubmitter


class RSSHubSubmitter(AsyncSubmitter, metaclass=abc.ABCMeta):
    logger = logging.getLogger("RSSHubSubmitter")

    @abc.abstractmethod
    async def parse_xml(self, xml: str) -> AsyncGenerator[dict, Any]:
        pass

    async def derive_tags(self, mission_content) -> List[str]:
        return ['rsshub']

    async def derive_matcher(self, mission_content) -> List[str]:
        return [mission_content['url']]

    async def create_missions(self, rsshub: str, **httpx_client_options):
        async with httpx.AsyncClient(**httpx_client_options) as client:
            response = await client.get(rsshub)
            async for mission_content in self.parse_xml(response.text):
                try:
                    matchers = await self.derive_matcher(mission_content)
                    await self.create_mission(mission_content, matchers)
                    tags = await self.derive_tags(mission_content)
                    if len(tags) > 0:
                        await self.add_tags(matchers, tags)
                except Exception as e:
                    self.logger.warning(f'create mission failed, error: {e}, traceback: {traceback.format_exc()}')


class RSSHubRootSubmitter(RSSHubSubmitter):

    async def parse_xml(self, xml: str) -> AsyncGenerator[dict, Any]:
        root = ElementTree.XML(xml)
        yield {
            'url': root.find('channel/link').text,
            'latest': [item.find('link').text for item in root.iter('item')][0]
        }


class RSSHubSubitemSubmitter(RSSHubSubmitter):

    async def parse_xml(self, xml: str) -> AsyncGenerator[dict, Any]:
        root = ElementTree.XML(xml)
        for item in root.find('channel').iter('item'):
            yield {'url': item.find('link').text}



import datetime
from sqlalchemy import (
    Column,
    Integer,
    Text,
    JSON,
    Boolean,
    DateTime,
    Interval,
    ForeignKey
)
from sqlalchemy.orm import relationship, Mapped
from .core import Base, Mission


class Attempt(Base):
    __tablename__ = "attempt"
    id = Column(Integer, primary_key=True, autoincrement=True, comment="Mission ID")
    handler = Column(Text, comment="Handler Name")
    create_time = Column(DateTime, default=datetime.datetime.now, comment="Attempt Start Time")
    last_update_time = Column(DateTime, default=datetime.datetime.now, onupdate=datetime.datetime.now, comment="Attempt Last Update Time")
    max_time_interval = Column(Interval, default=datetime.timedelta(seconds=1), comment="Attempt Update Time Interval")
    content = Column(JSON, default={}, comment="Mission Content at that time")
    success = Column(Boolean, default=False, comment="If this Attempt has succeed")

    # relationship
    mission_id = Column(Integer, ForeignKey("mission.id"), comment="Mission ID")
    mission: Mapped['Mission'] = relationship(Mission, backref="attempts")

    def __repr__(self):
        return f"Attempt(id={self.id}, handler={self.handler.__repr__()}, create_time={self.create_time.__repr__()}, last_update_time={self.last_update_time.__repr__()}, max_time_interval={self.max_time_interval.__repr__()}, content={self.content.__repr__()}, success={self.success}, mission_id={self.mission_id})"



import abc
import asyncio
import logging
import chardet
from typing import List, Callable
from missionpanel.handler import AsyncHandler
from missionpanel.orm.core import Mission
from missionpanel.orm.handler import Attempt


class SubprocessAsyncHandlerInterface(AsyncHandler, abc.ABC):
    def getLogger(self) -> logging.Logger:
        return logging.getLogger("SubprocessAsyncHandler")

    async def __readline_info(self, f):
        async for line in f:
            try:
                line = line.decode(chardet.detect(line)['encoding']).strip()
            except UnicodeDecodeError:
                line = line.strip()
            self.getLogger().info('stdout | %s' % line)

    async def __readline_debug(self, f):
        async for line in f:
            try:
                line = line.decode(chardet.detect(line)['encoding']).strip()
            except UnicodeDecodeError:
                line = line.strip()
            self.getLogger().info('stderr | %s' % line)

    @abc.abstractmethod
    async def construct_command(self, mission: Mission, attempt: Attempt) -> List[str]:
        raise NotImplementedError("Subclasses must implement construct_command")

    async def execute_mission(self, mission: Mission, attempt: Attempt) -> bool:
        proc = await asyncio.create_subprocess_exec(
            *(await self.construct_command(mission, attempt)),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        await asyncio.gather(self.__readline_info(proc.stdout), self.__readline_debug(proc.stderr))
        return_code = await proc.wait()
        self.getLogger().info('return | %d' % return_code)
        return return_code == 0



import abc
import asyncio
from typing import List

from missionpanel.orm import Mission, Attempt
from .handler import AsyncHandler, HandlerInterface


class ParallelAsyncHandler(AsyncHandler, abc.ABC):
    def __init__(self, n_parallel: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.task_queue = asyncio.Queue(n_parallel)
        for i in range(n_parallel):
            self.task_queue.put_nowait(i)
        self.task_dict = {}
        self.sem_report = asyncio.Semaphore(1)

    async def report_attempt(self, mission: Mission, attempt: Attempt):
        async with self.sem_report:
            await super().report_attempt(mission, attempt)

    async def run_all(self, tags: List[str]):
        async def task(mission: Mission, attempt: Attempt, id: int):
            attempt = await self.watchdog_mission(mission, attempt)
            await self.task_queue.put(id)
            return attempt
        while True:
            id = await self.task_queue.get()
            async with self.sem_report:
                mission = await self.get_mission(tags)
                if mission is None:
                    break
                attempt = HandlerInterface.create_attempt(self.session, mission, self.name, self.max_time_interval)
                self.task_dict[id] = asyncio.create_task(task(mission, attempt, id))
        for i in self.task_dict:
            await self.task_dict[i]
        self.task_dict = {}
